% Testing problem: Reconstruction of orientation distribution function (ODF).
%
% Matlab software for the paper "The regularizing effect of the Golub-Kahan 
% iterative bidiagonalization and revealing the noise level in the data", 
% by I. Hnetynkova, M. Plesinger, and Z. Strakos.
%
% Content.m ....... Table of content.
% csvd.m .......... Compact singular value decomposition, part of [1].
% goka_b.m ........ Golub-Kahan bidiagonalization; it is a modified routine 
%                   lanc_b.m from [1]. (Note that, if the bidiagonaliztion is
%                   computed with Householder reorthogonalization, the routine
%                   app_hh.m from [1] is needed, it is not a part of this pack.)
% ODFdata.mat ..... Contains all data for the problem:
%                   the system matrix "A", 
%                   the exact right-hand side "b" (also in "bexact"), and
%                   the exact solution "x" (also rearranged in 3D array "X3D"). 
% testODF.m ....... Launches the testing problem!
% 
% [1] P. C. Hansen -- Regularization Tools, Version 4.0 for MATLAB 7.3.

% Author of this code: P. C. Hansen; modified by M. Plesinger, April, 2009.

clear;

% Number of iterations of the Golub-Kahan bidiagonalization ...
kmax = 200;

% ... without reorthogonalization.
reorth = 0;

% Loading the sytem matrix, the exact right-hand side, and the exact solution.
fprintf('\nLoading data (A, b_exact, and x_exact) ...........');
load ODFdata
x_exact = x;
[m,n] = size(A);
fprintf(' OK\n');

% The exact right-hand.
fprintf('Constructing the noisy right-hand side ...........');
b_exact = b;

% The additional white-noise and the noisy right-hand side.
delta_noise = 3e-3;
b_noise = randn(m,1); 
b_noise = b_noise/norm(b_noise);
b_noise = norm(b_exact) * b_noise * delta_noise;
b = b_exact + b_noise;
beta = norm(b);
fprintf(' OK\n');

% Run the Golub-Kahan bidiagonalization.
fprintf('Computing the Golub-Kahan bidiagonalization ......');
[S,L,W] = goka_b(A,b,kmax,reorth);
fprintf(' OK\n');

% Initialization: noise level estimator, approx. soltuions, and error history. 
fprintf('Estimating the noise level, and computing\n');
fprintf('the LSQR approximations of the exact solution ....');
nl_estimate = zeros(kmax,1);
X = zeros(n,kmax);
err = zeros(kmax,1);

% Computes the SVD of L_k and approximates the solution, for k = 1,...,kmax.
for k = 1:kmax,

    % The SVD components has standard ordeding, here. Please note that in the
    % paper is used inverse ordering. Thus, we look for the first component
    % of the left singular vector corresponding to Theta(k,k) here.
    [P,Theta,Q] = csvd(L(1:k,1:k));
    nl_estimate(k) = P(1,k);
    
    % LSQR approximation of the solution, obtained using backslash operator!!!
    e1 = zeros(k+1,1); 
    e1(1) = 1;
    X(:,k) = beta*W(:,1:k)*(L(1:k+1,1:k)\e1);
    err(k,1) = norm(X(:,k)-x_exact);

end
[minerr,Idx] = min(err);
fprintf(' OK\n');

% Plotting graphs ...
fprintf('Plotting graphs ..................................');
figure(1); clf;

% ... of the estimator, ....
subplot(2,1,1);
h = semilogy(1:kmax,abs(nl_estimate),'k-o',[1,kmax],delta_noise*[1,1],'r-.');
set(h(1),'LineWidth',2);
title(['| ( p_1^{(k)} , e_1 ) |,   || b^{noise} ||_2 / || b^{exact} ||_2 = ', num2str(norm(b_noise)/norm(b_exact))]);
legend('| ( p_1^{(k)} , e_1 ) |','\delta_{noise}');

% ... and the error history.
subplot(2,1,2);
h = semilogy(1:kmax,err,'k-',Idx,err(Idx),'ro');
set(h(2),'LineWidth',2);
set(h(2),'MarkerSize',12);
axis([0,kmax,err(Idx)*0.9,max(err)/0.9]);
title('error history');
legend('|| x_k^{LSQR} - x^{exact} ||',sprintf('the minimal error, k = %d',Idx));
fprintf(' OK\n\n');

